/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* main.cpp
   Gabriel M. Beddingfield
   2008-02-10

   Driver to test different tap tempo models.

 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <deque>
#include <cmath>
#include <ctime>
#include <sys/time.h>

#include <alsa/asoundlib.h>
#include <alsa/seq.h>

// Using my epoch, I can take the 12-byte timeval struct
// and stick it in an 8-byte unsigned long long.
// Jan 1, 2008, in unix time
const unsigned long long MYEPOCH = 1199167200L;
// this test should work fine until at least 2070.

typedef unsigned long long usecs_t;  // microseconds

static inline usecs_t get_cycles (void)
{
    usecs_t rv;
    timeval tv;
    gettimeofday(&tv, 0);
    rv = (tv.tv_sec - MYEPOCH)*1000000 + tv.tv_usec;
    return rv;
}

using namespace std;

int main(int argc, char* argv[])
{
    snd_seq_t* seq;
    int seq_port;
    ofstream ofile;

    cout << "Tap Tempo Experiments (Data Collection)" << endl
	 << "2008-02-10 Gabriel M. Beddingfield" << endl
	 << endl;

    if (argc != 2) {
        cout << endl
             << "usage: " << argv[0] << " outfile" << endl
             << endl;
        exit(0);
    } else {
        ofile.open(argv[1]);
        if (ofile.fail()) {
            cout << "could not open file " << argv[1] << "for writing" << endl
                 << "...aborting." << endl
                 << endl;
            exit(0);
        }
    }

    cout << "Opening MIDI port" << endl;

    snd_seq_open(&seq, "default", SND_SEQ_OPEN_INPUT, 0);
    seq_port = snd_seq_create_simple_port(
	seq,
	"Tap Tempo Experiment In",
        SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
        SND_SEQ_PORT_TYPE_MIDI_GENERIC|SND_SEQ_PORT_TYPE_APPLICATION
        |SND_SEQ_PORT_TYPE_SOFTWARE);

    cout << "Listening for taps on channel 12 (b)." << endl
	 << "Send 0xbb 0x32 0x7f (CC/12 50 127) to register a tap." << endl
	 << "Send 0xbb 0x31 0x00 (CC/12 49 0) to exit application."
	 << endl;

    const char* speeds[] = { "slowly", "medium", "fast" };
    const unsigned nspeeds =  3;  // speed[0..(nspeeds-1)]
    const unsigned taps[] = { 3, 4, 5, 6 };
    const unsigned ntaps = 4;  // taps[0..(ntaps-1)]
    unsigned ntests = ntaps * nspeeds;
    unsigned test = 0;
    snd_seq_event_t* ev;
    usecs_t now;
    bool not_done = true;
    while(not_done) {
	int getting_taps = taps[test/nspeeds];

	snd_seq_drop_input(seq);
	cout << "Please give " << getting_taps << " taps "
	     << speeds[test%nspeeds] << endl;

        ofile << getting_taps << " " << speeds[test%nspeeds] << ":";

	while(getting_taps > 0) {
	    snd_seq_event_input(seq, &ev);
	    now = get_cycles();  // hurry!

	    // Filters
	    if(ev->type != SND_SEQ_EVENT_CONTROLLER)
		continue;
	    if(ev->data.control.channel != 0x0b)
		continue;
	    if(ev->data.control.param == 0x31)
		exit(0);
	    if((ev->data.control.param != 0x32)
	       && (ev->data.control.value & 0x40))
		continue;

	    cout << "Tap @ " << dec << now << endl;
            ofile << " " << dec << now;

	    // Register taps.
	    --getting_taps;
	}
        ofile << endl;

	++test;
	if(test >= ntests) not_done = false;
    }

    ofile.close();
    cout << "Closing MIDI port" << endl;
    snd_seq_delete_simple_port(seq, seq_port);
    return 0;
}
