/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version                                                               *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
/* MidiPort.cpp
 * Copyright(c) 2008 by Gabriel M. Beddingfield <gabriel@teuton.org>
 */

#ifdef JACK_MIDI_SUPPORT

#include "MidiInPort.hpp"
#include "Client.hpp"
#include <cassert>
#include <cstdlib> // free()
#include <cerrno> // EEXIST for jack_connect()

/*********************************************************
 * Notes on JACK versions with respect to MIDI
 *
 * First, jack/version.h defines the macro JACK_PROTOCOL_VERSION
 * to an integer... and this is *supposed* to be the reliable indicator
 * of our API.  However, this value is not reliable with respect
 * to JACK MIDI.  The following table describes the different
 * JACK versions and how they relate to the MIDI API.
 *
 * SVN  JACK Version PROTOCOL Milestone
 * ---- ------------ -------- -----------------------------
 *  847   0.99.37       15    Protocol set to 15
 *  ---   0.101.1       15    A public release
 *  945   0.102.0       15    Original MIDI API.
 *  968   0.102.10      16    Protocol set to 16
 *  ---   0.102.20      16    A public release
 *  998   0.102.27      16    MIDI API Changed:
 *                            jack_midi_port_get_info() replaced by
 *                              jack_midi_get_event_count()
 *                            jack_midi_reset_new_port() removed
 *  ---   0.103.0       16    A public release
 * 1030   0.103.0       19    Protocol set to 19
 * 1032   0.104.0       20    Protocol set to 20
 * 1035   0.105.0       20    MIDI API Changed:
 *                            The nframes paramter removed from these:
 *                              jack_midi_get_event_count()
 *                              jack_midi_event_get()
 *                              jack_midi_clear_buffer()
 *                              jack_midi_max_event_size()
 *                              jack_midi_event_reserve()
 *                              jack_midi_event_write()
 *                              jack_midi_get_lost_event_count()
 * 1046   0.107.0       21    Protocol set to 21
 *    ....
 *  ---   0.109.0       22    A public release
 *
 * It's also worth noting that the version numbers tend to change
 * with every commit.  However, the protocol version changes have
 * odd overlaps (with respect to MIDI).
 *
 * To manage this while the API stabilizes, the following macros
 * should be set by the build system:
 *
 * JACK_MIDI_SUPPORT - Include JACK MIDI support
 *
 * Then, there must be one and only one of the following set:
 *
 *    JACK_MIDI_0_102_0  - Conforms to API of 0.102.0
 *    JACK_MIDI_0_102_27 - Conforms to API of 0.102.27
 *    JACK_MIDI_0_105_0  - Conforms to API of 0.105.0
 *
 ********************************************************
 */

#if defined(JACK_MIDI_0_102_0) && defined(JACK_MIDI_0_102_27)
#  error "JACK_MIDI_0_102_0 and JACK_MIDI_0_102_27 both defined -- only ONE should be."
#endif
#if defined(JACK_MIDI_0_102_0) && defined(JACK_MIDI_0_105_0)
#  error "JACK_MIDI_0_102_0 and JACK_MIDI_0_105_0 both defined -- only ONE should be."
#endif
#if defined(JACK_MIDI_0_102_27) && defined(JACK_MIDI_0_105_0)
#  error "JACK_MIDI_0_102_27 and JACK_MIDI_0_105_0 both defined -- only ONE should be."
#endif

#if !defined(JACK_MIDI_0_102_0) && !defined(JACK_MIDI_0_102_27) && !defined(JACK_MIDI_0_105_0)
#  error "One of these must be defined:  JACK_MIDI_0_102_0, JACK_MIDI_0_102_27, JACK_MIDI_0_105_0"
#endif

// Macro to help readability for API change where nframes parameter
// is eliminated.
#ifdef JACK_MIDI_0_105_0
#  define NFPARAM(x)
#else
#  define NFPARAM(x) , (x)
#endif

using namespace std;
using namespace HiJack;

MidiInPort::MidiInPort(const std::string& name) :
    Port(name, Port::Midi)
{
}

MidiInPort::MidiInPort(Client *client,
                       const std::string& name) :
    Port(client, name, Port::Midi, JackPortIsInput)
{
}

MidiInPort::~MidiInPort()
{
}

void MidiInPort::handleMidi(const jack_midi_event_t& event)
{
    // For midi_commands, it's the programmer's responsibility
    // to ensure:
    //
    // a) The MSB is 1 (i.e. 1xxx xxxx)
    // b) the LSW is 0 (i.e. xxxx 0000)
    // c) That all 8 possible MSW's are present in the enum (0x8 thru 0xF)
    //
    // If not, casting ints to enums will fail:  either giving you
    // a runtime error or unexpected results.
    typedef enum {
        UNKNOWN = 0x00,
        NOTE_OFF = 0x80,
        NOTE_ON = 0x90,
        POLYPHONIC_KEY_PRESSURE = 0xA0,
        CONTROL_CHANGE = 0xB0,
        PROGRAM_CHANGE = 0xC0,
        CHANNEL_PRESSURE = 0xD0,
        PITCH_WHEEL = 0xE0,
        SYSTEM_EXCLUSIVE = 0xF0
    } midi_commands;
    // For sysex_messages, it's the programmer's responsibility
    // to ensure:
    //
    // a) the MSW is 0 (i.e. 0000 xxxx)
    // b) That all 16 possible LSW's are present in the enum (0x0 thru 0xF)
    //
    // If not, casting ints to enums will fail:  either giving you
    // a runtime error or unexpected results.
    typedef enum {
        SYSEX_START = 0x00,
        MTC_QUARTER_FRAME = 0x01,
        SONG_POS = 0x02,
        SONG_SELECT = 0x03,
        UNDEFINED_04 = 0x04,
        UNDEFINED_05 = 0x05,
        TUNE_REQ = 0x06,
        SYSEX_END = 0x07,
        CLOCK = 0x08,
        UNDEFINED_09 = 0x09,
        SONG_START = 0x0A,
        SONG_CONT = 0x0B,
        SONG_STOP = 0x0C,
        UNDEFINED_0D = 0x0D,
        ACTIVE_SENSING = 0x0E,
        SYSTEM_RESET = 0x0F
    } sysex_messages;
        

    midi_commands status;
    unsigned char tmp;
    unsigned char chan;
    uint16_t data16;
    #warning "Sequence type std::list<uint8_t> is probably not realtime safe."
    std::list<uint8_t> sysex_data;

    if (event.size == 0)
        return;

    // Test the first byte... which should always have the "status
    // bit" (bit 7) as a 1 since JACK supplies normalized events.
    tmp = event.buffer[0] & 0xF0;
    if (0x80 & tmp) {
        assert( (tmp & 0x8F) == 0x80 );
        status = (midi_commands)tmp;
    } else {
        status = UNKNOWN;
    }

    chan = event.buffer[0] & 0x0F;

    switch(status) {
    case UNKNOWN:
        break;
    case NOTE_ON:
        handleNoteOn(chan, event.buffer[1], event.buffer[2], event.time);
        break;
    case NOTE_OFF:
        handleNoteOff(chan, event.buffer[1], event.buffer[2], event.time);
        break;
    case POLYPHONIC_KEY_PRESSURE:
        handlePolyphonicKeyPressure(chan, event.buffer[1],
                                    event.buffer[2], event.time);
        break;
    case CONTROL_CHANGE:
        handleControlChange(chan, event.buffer[1], event.buffer[2], event.time);
        break;
    case PROGRAM_CHANGE:
        handleProgramChange(chan, event.buffer[1], event.time);
        break;
    case CHANNEL_PRESSURE:
        handleChannelPressure(chan, event.buffer[1], event.time);
        break;
    case PITCH_WHEEL:
        data16 = (event.buffer[2] << 7) | event.buffer[1];
        handlePitchWheel(chan, data16, event.time);
        break;
    case SYSTEM_EXCLUSIVE:
        assert( ((event.buffer[0] & 0x0F) & 0xF0) == 0 );
        switch (event.buffer[0] & 0x0F) {
        case SYSEX_START:
            sysex_data.insert(sysex_data.end(), event.buffer+2, event.buffer+event.size);
            handleSysCommExclusive(event.buffer[1], sysex_data, event.time);
            break;
        case MTC_QUARTER_FRAME:
            handleSysCommMTCQuarterFrame(event.buffer[1], event.time);
            break;
        case SONG_POS:
            data16 = (event.buffer[2] << 7) | event.buffer[1];
            handleSysCommSongPosition(data16, event.time);
            break;
        case SONG_START:
            handleSysRealStart(event.time);
            break;
        case SONG_CONT:
            handleSysRealContinue(event.time);
            break;
        case SONG_STOP:
            handleSysRealStop(event.time);
            break;
        case SONG_SELECT:
            handleSysCommSongSelect(event.buffer[1], event.time);
            break;
        case TUNE_REQ:
            handleSysCommTuneRequest(event.time);
            break;
        case CLOCK:
            handleSysRealTimingClock(event.time);
            break;
        case ACTIVE_SENSING:
            handleSysRealActiveSensing(event.time);
            break;
        case SYSTEM_RESET:
            handleSysRealReset(event.time);
            break;
            // Following are unhandled.
        case SYSEX_END:
            // JACK should never give us this one.
        case UNDEFINED_04:
        case UNDEFINED_05:
        case UNDEFINED_09:
        case UNDEFINED_0D:
            break;
        default:
            assert(false);  // Should not reach this line
        }
        break;
    default:
        assert(false); // Should not reach this line
    }
}

// This function must be realtime safe.  It will be called from
// the JACK process callback.
int MidiInPort::processCallback(jack_nframes_t nframes)
{
    jack_port_t* port = ref();
    if (!port) return 0;

    jack_nframes_t event_ct, event_pos;
    jack_midi_event_t jack_event;

    void* port_buf = jack_port_get_buffer(port, nframes);
#ifdef JACK_MIDI_0_102_0
    event_ct = jack_midi_port_get_info(port_buf, nframes)->event_count;
#else
    event_ct = jack_midi_get_event_count(port_buf NFPARAM(nframes));
#endif

    for ( event_pos=0 ; event_pos<event_ct ; ++event_pos ) {
        if (jack_midi_event_get(&jack_event,
                                port_buf,
                                event_pos
                                NFPARAM(nframes))) {
            break;
        }
        handleMidi(jack_event);
    }
    return 0;
}

// std::vector<QString> MidiInPort::getOutputPortList(void)
// {
//     return Client::get_instance()->getMidiOutputPortList();
// }

/////////////////////////////////////////////////////////////////
// DEFAULT EVENT HANDLERS (THEY DO NOTHING)
// NOTE THAT THESE *ALL* MUST BE REALTIME SAFE
/////////////////////////////////////////////////////////////////

void MidiInPort::handleNoteOn(byte /*chan*/,
                              byte /*note*/,
                              byte /*velocity*/,
                              jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleNoteOff(byte /*chan*/,
                               byte /*note*/,
                               byte /*velocity*/,
                               jack_nframes_t /*frame*/)
{
}

void MidiInPort::handlePolyphonicKeyPressure(
    byte /*chan*/,
    byte /*note*/,
    byte /*value*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleControlChange(
    byte /*chan*/,
    byte /*ctrl_no*/,
    byte /*value*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleProgramChange(
    byte /*chan*/,
    byte /*program*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleChannelPressure(
    byte /*chan*/,
    byte /*value*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handlePitchWheel(byte /*chan*/,
                                  uint16_t /*value*/,
                                  jack_nframes_t /*frame*/)
{
}

// System Common
void MidiInPort::handleSysCommExclusive(
    byte /*vendid*/,
    const std::list<byte>& /*data*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysCommMTCQuarterFrame(
    byte /*value*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysCommSongPosition(
    uint16_t /*value*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysCommSongSelect(
    byte /*song*/,
    jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysCommTuneRequest(jack_nframes_t /*frame*/)
{
}

// System Realtime
void MidiInPort::handleSysRealTimingClock(jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysRealStart(jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysRealContinue(jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysRealStop(jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysRealActiveSensing(jack_nframes_t /*frame*/)
{
}

void MidiInPort::handleSysRealReset(jack_nframes_t /*frame*/)
{
}

#endif JACK_MIDI_SUPPORT
