#ifndef __TAPTEMPOINTERFACE_HPP__
#define __TAPTEMPOINTERFACE_HPP__
/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TapTempoInterface.hpp
   Gabriel M. Beddingfield
   2008-08-16

 */

#include <string>

namespace InConcert
{

    /*
     * class TapTempoInterface
     *
     * This class provides a portable interface to different tap
     * tempo models and time-measuring types.
     *
     */
    template <typename ScalarT, typename TempoT = double>
    class TapTempoInterface
    {
    public:
        virtual ~TapTempoInterface() { }

	/**
	 * user-accessible type info.
	 */
	typedef ScalarT scalar_type;
	typedef TempoT tempo_type;

	/**
	 * name()
	 *
	 * Returns a user recognizable name for this tap
	 * tempo model. (E.g. "Least Squares" or "Rolling Average").
	 */
	virtual const std::string& name(void) const = 0;

	/**
	 * Get/Set the number of taps required to calculate
	 * the next beat/tempo.
	 */
        virtual unsigned getTapsRequired(void) const = 0;
        virtual void setTapsRequired(unsigned t) = 0;

	/**
	 * Get/Set the rate multiplier for returning the "Tempo."
	 *
	 * For example, if the current sample rate is 44,100 Hz,
	 * then 120 beats/min is a sample interval of 22,050.
	 * However, there are 60 * 44,100 = 2,646,000 samples in
	 * a minute.  So if the taps are spaced 22,050 apart
	 * and the rate multiplier is 2,646,000 then the
	 * tempo returned will be 2,646,000 / 22,050 = 120.0
	 *
	 */
        virtual const TempoT& getRateMultiplier(void) const = 0;
        virtual void setRateMultiplier(TempoT rate) = 0;

	/**
	 * Get/Set the number of samples that, once elapsed,
	 * causes this tap tempo model to invalidate.  (E.g.
	 * when 2 seconds pass)
	 */
        virtual const ScalarT& getTimeout(void) const = 0;
        virtual void setTimeout(ScalarT t) = 0;

	/**
	 * Start the tap tempo model over.
	 */
        virtual void invalidate(void) = 0;
    
	/*
	 * tap()
	 *
	 * Register a tap to define a new tempo and downbeat sync.  Supply
	 * a reference time 'tap_' (e.g. frame number).
	 *
	 * Returns 0 if the model is not able to predict the next beat.
	 * If the model is able to predict the next beat, then will return
	 * a non-zero value.  If it returns non-zero, then the tempo may
	 * be obtained by calling getTempo().
	 *
	 */
        virtual ScalarT tap(ScalarT tap_) = 0;

	/**
	 * getTempoRate(), getTempoInterval()
	 *
	 * If tap() returns a non-zero, then these functions will
	 * return the tempo determined by the model.  Otherwise,
	 * returns 0.  getTempoInterval() gives the number of ScalarT
	 * measures between beats.  getTempoRate() returns
	 * getRateMultiplier()/getTempoInterval().
	 */
	virtual const TempoT& getTempoRate(void) const = 0;
	virtual const ScalarT& getTempoInterval(void) const = 0;
    }; // class TapTempoInterface
    
} // namespace InConcert
    
#endif // __TAPTEMPOINTERFACE_HPP__
